#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib
import os

from bs4 import BeautifulSoup
from sphinx import errors
from sphinx.util import logging

LOG = logging.getLogger(__name__)

# NOTE: We do this because I can't find any way to pass "-v"
# into sphinx-build through pbr...
DEBUG = True if os.getenv("ENFORCER_DEBUG") else False

WRITTEN_METHODS = set()


class EnforcementError(errors.SphinxError):
    """A mismatch between what exists and what's documented"""
    category = "Enforcer"


def get_proxy_methods():
    """Return a set of public names on all proxies"""
    names = ["openstack.baremetal.v1._proxy",
             "openstack.clustering.v1._proxy",
             "openstack.block_storage.v2._proxy",
             "openstack.compute.v2._proxy",
             "openstack.database.v1._proxy",
             "openstack.identity.v2._proxy",
             "openstack.identity.v3._proxy",
             "openstack.image.v1._proxy",
             "openstack.image.v2._proxy",
             "openstack.key_manager.v1._proxy",
             "openstack.load_balancer.v2._proxy",
             "openstack.message.v2._proxy",
             "openstack.network.v2._proxy",
             "openstack.object_store.v1._proxy",
             "openstack.orchestration.v1._proxy",
             "openstack.workflow.v2._proxy"]

    modules = (importlib.import_module(name) for name in names)

    methods = set()
    for module in modules:
        # We're not going to use the Proxy for anything other than a `dir`
        # so just pass a dummy value so we can create the instance.
        instance = module.Proxy("")
        # We only document public names
        names = [name for name in dir(instance) if not name.startswith("_")]

        good_names = [module.__name__ + ".Proxy." + name for name in names]
        methods.update(good_names)

    return methods


def page_context(app, pagename, templatename, context, doctree):
    """Handle html-page-context-event

    This event is emitted once the builder has the contents to create
    an HTML page, but before the template is rendered. This is the point
    where we'll know what documentation is going to be written, so
    gather all of the method names that are about to be included
    so we can check which ones were or were not processed earlier
    by autodoc.
    """
    if "users/proxies" in pagename:
        soup = BeautifulSoup(context["body"], "html.parser")
        dts = soup.find_all("dt")
        ids = [dt.get("id") for dt in dts]

        written = 0
        for id in ids:
            if id is not None and "_proxy.Proxy" in id:
                WRITTEN_METHODS.add(id)
                written += 1

        if DEBUG:
            LOG.info("ENFORCER: Wrote %d proxy methods for %s" % (
                     written, pagename))


def build_finished(app, exception):
    """Handle build-finished event

    This event is emitted once the builder has written all of the output.
    At this point we just compare what we know was written to what we know
    exists within the modules and share the results.

    When enforcer_warnings_as_errors=True in conf.py, this method
    will raise EnforcementError on any failures in order to signal failure.
    """
    all_methods = get_proxy_methods()

    LOG.info("ENFORCER: %d proxy methods exist" % len(all_methods))
    LOG.info("ENFORCER: %d proxy methods written" % len(WRITTEN_METHODS))
    missing = all_methods - WRITTEN_METHODS

    missing_count = len(missing)
    LOG.info("ENFORCER: Found %d missing proxy methods "
             "in the output" % missing_count)

    # TODO(shade) This is spewing a bunch of content for missing thing that
    # are not actually missing. Leave it as info rather than warn so that the
    # gate doesn't break ... but we should figure out why this is broken and
    # fix it.
    # We also need to deal with Proxy subclassing keystoneauth.adapter.Adapter
    # now - some of the warnings come from Adapter elements.
    for name in sorted(missing):
        LOG.info("ENFORCER: %s was not included in the output" % name)

    if app.config.enforcer_warnings_as_errors and missing_count > 0:
        raise EnforcementError(
            "There are %d undocumented proxy methods" % missing_count)


def setup(app):
    app.add_config_value("enforcer_warnings_as_errors", False, "env")

    app.connect("html-page-context", page_context)
    app.connect("build-finished", build_finished)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import mock

from openstack.compute.v2 import _proxy
from openstack.compute.v2 import availability_zone as az
from openstack.compute.v2 import extension
from openstack.compute.v2 import flavor
from openstack.compute.v2 import hypervisor
from openstack.compute.v2 import image
from openstack.compute.v2 import keypair
from openstack.compute.v2 import limits
from openstack.compute.v2 import server
from openstack.compute.v2 import server_group
from openstack.compute.v2 import server_interface
from openstack.compute.v2 import server_ip
from openstack.compute.v2 import service
from openstack.tests.unit import test_proxy_base


class TestComputeProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super(TestComputeProxy, self).setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_extension_find(self):
        self.verify_find(self.proxy.find_extension, extension.Extension)

    def test_extensions(self):
        self.verify_list_no_kwargs(self.proxy.extensions, extension.Extension)

    def test_flavor_create(self):
        self.verify_create(self.proxy.create_flavor, flavor.Flavor)

    def test_flavor_delete(self):
        self.verify_delete(self.proxy.delete_flavor, flavor.Flavor, False)

    def test_flavor_delete_ignore(self):
        self.verify_delete(self.proxy.delete_flavor, flavor.Flavor, True)

    def test_flavor_find(self):
        self.verify_find(self.proxy.find_flavor, flavor.Flavor)

    def test_flavor_get(self):
        self.verify_get(self.proxy.get_flavor, flavor.Flavor)

    def test_flavors_detailed(self):
        self.verify_list(self.proxy.flavors, flavor.FlavorDetail,
                         method_kwargs={"details": True, "query": 1},
                         expected_kwargs={"query": 1})

    def test_flavors_not_detailed(self):
        self.verify_list(self.proxy.flavors, flavor.Flavor,
                         method_kwargs={"details": False, "query": 1},
                         expected_kwargs={"query": 1})

    def test_image_delete(self):
        self.verify_delete(self.proxy.delete_image, image.Image, False)

    def test_image_delete_ignore(self):
        self.verify_delete(self.proxy.delete_image, image.Image, True)

    def test_image_find(self):
        self.verify_find(self.proxy.find_image, image.Image)

    def test_image_get(self):
        self.verify_get(self.proxy.get_image, image.Image)

    def test_images_detailed(self):
        self.verify_list(self.proxy.images, image.ImageDetail,
                         method_kwargs={"details": True, "query": 1},
                         expected_kwargs={"query": 1})

    def test_images_not_detailed(self):
        self.verify_list(self.proxy.images, image.Image,
                         method_kwargs={"details": False, "query": 1},
                         expected_kwargs={"query": 1})

    def test_keypair_create(self):
        self.verify_create(self.proxy.create_keypair, keypair.Keypair)

    def test_keypair_delete(self):
        self.verify_delete(self.proxy.delete_keypair, keypair.Keypair, False)

    def test_keypair_delete_ignore(self):
        self.verify_delete(self.proxy.delete_keypair, keypair.Keypair, True)

    def test_keypair_find(self):
        self.verify_find(self.proxy.find_keypair, keypair.Keypair)

    def test_keypair_get(self):
        self.verify_get(self.proxy.get_keypair, keypair.Keypair)

    def test_keypairs(self):
        self.verify_list_no_kwargs(self.proxy.keypairs, keypair.Keypair)

    def test_limits_get(self):
        self.verify_get(self.proxy.get_limits, limits.Limits, value=[])

    def test_server_interface_create(self):
        self.verify_create(self.proxy.create_server_interface,
                           server_interface.ServerInterface,
                           method_kwargs={"server": "test_id"},
                           expected_kwargs={"server_id": "test_id"})

    def test_server_interface_delete(self):
        self.proxy._get_uri_attribute = lambda *args: args[1]

        interface_id = "test_interface_id"
        server_id = "test_server_id"
        test_interface = server_interface.ServerInterface(id=interface_id)
        test_interface.server_id = server_id

        # Case1: ServerInterface instance is provided as value
        self._verify2("openstack.proxy.Proxy._delete",
                      self.proxy.delete_server_interface,
                      method_args=[test_interface],
                      method_kwargs={"server": server_id},
                      expected_args=[server_interface.ServerInterface],
                      expected_kwargs={"server_id": server_id,
                                       "port_id": interface_id,
                                       "ignore_missing": True})

        # Case2: ServerInterface ID is provided as value
        self._verify2("openstack.proxy.Proxy._delete",
                      self.proxy.delete_server_interface,
                      method_args=[interface_id],
                      method_kwargs={"server": server_id},
                      expected_args=[server_interface.ServerInterface],
                      expected_kwargs={"server_id": server_id,
                                       "port_id": interface_id,
                                       "ignore_missing": True})

    def test_server_interface_delete_ignore(self):
        self.proxy._get_uri_attribute = lambda *args: args[1]
        self.verify_delete(self.proxy.delete_server_interface,
                           server_interface.ServerInterface, True,
                           method_kwargs={"server": "test_id"},
                           expected_args=[server_interface.ServerInterface],
                           expected_kwargs={"server_id": "test_id",
                                            "port_id": "resource_or_id"})

    def test_server_interface_get(self):
        self.proxy._get_uri_attribute = lambda *args: args[1]

        interface_id = "test_interface_id"
        server_id = "test_server_id"
        test_interface = server_interface.ServerInterface(id=interface_id)
        test_interface.server_id = server_id

        # Case1: ServerInterface instance is provided as value
        self._verify2('openstack.proxy.Proxy._get',
                      self.proxy.get_server_interface,
                      method_args=[test_interface],
                      method_kwargs={"server": server_id},
                      expected_args=[server_interface.ServerInterface],
                      expected_kwargs={"port_id": interface_id,
                                       "server_id": server_id})

        # Case2: ServerInterface ID is provided as value
        self._verify2('openstack.proxy.Proxy._get',
                      self.proxy.get_server_interface,
                      method_args=[interface_id],
                      method_kwargs={"server": server_id},
                      expected_args=[server_interface.ServerInterface],
                      expected_kwargs={"port_id": interface_id,
                                       "server_id": server_id})

    def test_server_interfaces(self):
        self.verify_list(self.proxy.server_interfaces,
                         server_interface.ServerInterface,
                         method_args=["test_id"],
                         expected_kwargs={"server_id": "test_id"})

    def test_server_ips_with_network_label(self):
        self.verify_list(self.proxy.server_ips, server_ip.ServerIP,
                         method_args=["test_id"],
                         method_kwargs={"network_label": "test_label"},
                         expected_kwargs={"server_id": "test_id",
                                          "network_label": "test_label"})

    def test_server_ips_without_network_label(self):
        self.verify_list(self.proxy.server_ips, server_ip.ServerIP,
                         method_args=["test_id"],
                         expected_kwargs={"server_id": "test_id",
                                          "network_label": None})

    def test_server_create_attrs(self):
        self.verify_create(self.proxy.create_server, server.Server)

    def test_server_delete(self):
        self.verify_delete(self.proxy.delete_server, server.Server, False)

    def test_server_delete_ignore(self):
        self.verify_delete(self.proxy.delete_server, server.Server, True)

    def test_server_force_delete(self):
        self._verify("openstack.compute.v2.server.Server.force_delete",
                     self.proxy.delete_server,
                     method_args=["value", False, True])

    def test_server_find(self):
        self.verify_find(self.proxy.find_server, server.Server)

    def test_server_get(self):
        self.verify_get(self.proxy.get_server, server.Server)

    def test_servers_detailed(self):
        self.verify_list(self.proxy.servers, server.Server,
                         method_kwargs={"details": True,
                                        "changes_since": 1, "image": 2},
                         expected_kwargs={"changes_since": 1, "image": 2,
                                          "base_path": "/servers/detail"})

    def test_servers_not_detailed(self):
        self.verify_list(self.proxy.servers, server.Server,
                         method_kwargs={"details": False,
                                        "changes_since": 1, "image": 2},
                         expected_kwargs={"changes_since": 1, "image": 2})

    def test_server_update(self):
        self.verify_update(self.proxy.update_server, server.Server)

    def test_server_wait_for(self):
        value = server.Server(id='1234')
        self.verify_wait_for_status(
            self.proxy.wait_for_server,
            method_args=[value],
            expected_args=[value, 'ACTIVE', ['ERROR'], 2, 120])

    def test_server_resize(self):
        self._verify("openstack.compute.v2.server.Server.resize",
                     self.proxy.resize_server,
                     method_args=["value", "test-flavor"],
                     expected_args=["test-flavor"])

    def test_server_confirm_resize(self):
        self._verify("openstack.compute.v2.server.Server.confirm_resize",
                     self.proxy.confirm_server_resize,
                     method_args=["value"])

    def test_server_revert_resize(self):
        self._verify("openstack.compute.v2.server.Server.revert_resize",
                     self.proxy.revert_server_resize,
                     method_args=["value"])

    def test_server_rebuild(self):
        id = 'test_image_id'
        image_obj = image.Image(id='test_image_id')

        # Case1: image object is provided
        # NOTE: Inside of Server.rebuild is where image_obj gets converted
        # to an ID instead of object.
        self._verify('openstack.compute.v2.server.Server.rebuild',
                     self.proxy.rebuild_server,
                     method_args=["value", "test_server", "test_pass"],
                     method_kwargs={"metadata": {"k1": "v1"},
                                    "image": image_obj},
                     expected_args=["test_server", "test_pass"],
                     expected_kwargs={"metadata": {"k1": "v1"},
                                      "image": image_obj})

        # Case2: image name or id is provided
        self._verify('openstack.compute.v2.server.Server.rebuild',
                     self.proxy.rebuild_server,
                     method_args=["value", "test_server", "test_pass"],
                     method_kwargs={"metadata": {"k1": "v1"},
                                    "image": id},
                     expected_args=["test_server", "test_pass"],
                     expected_kwargs={"metadata": {"k1": "v1"},
                                      "image": id})

    def test_add_fixed_ip_to_server(self):
        self._verify("openstack.compute.v2.server.Server.add_fixed_ip",
                     self.proxy.add_fixed_ip_to_server,
                     method_args=["value", "network-id"],
                     expected_args=["network-id"])

    def test_fixed_ip_from_server(self):
        self._verify("openstack.compute.v2.server.Server.remove_fixed_ip",
                     self.proxy.remove_fixed_ip_from_server,
                     method_args=["value", "address"],
                     expected_args=["address"])

    def test_floating_ip_to_server(self):
        self._verify("openstack.compute.v2.server.Server.add_floating_ip",
                     self.proxy.add_floating_ip_to_server,
                     method_args=["value", "floating-ip"],
                     expected_args=["floating-ip"],
                     expected_kwargs={'fixed_address': None})

    def test_add_floating_ip_to_server_with_fixed_addr(self):
        self._verify("openstack.compute.v2.server.Server.add_floating_ip",
                     self.proxy.add_floating_ip_to_server,
                     method_args=["value", "floating-ip", 'fixed-addr'],
                     expected_args=["floating-ip"],
                     expected_kwargs={'fixed_address': 'fixed-addr'})

    def test_remove_floating_ip_from_server(self):
        self._verify("openstack.compute.v2.server.Server.remove_floating_ip",
                     self.proxy.remove_floating_ip_from_server,
                     method_args=["value", "address"],
                     expected_args=["address"])

    def test_server_backup(self):
        self._verify("openstack.compute.v2.server.Server.backup",
                     self.proxy.backup_server,
                     method_args=["value", "name", "daily", 1],
                     expected_args=["name", "daily", 1])

    def test_server_pause(self):
        self._verify("openstack.compute.v2.server.Server.pause",
                     self.proxy.pause_server,
                     method_args=["value"])

    def test_server_unpause(self):
        self._verify("openstack.compute.v2.server.Server.unpause",
                     self.proxy.unpause_server,
                     method_args=["value"])

    def test_server_suspend(self):
        self._verify("openstack.compute.v2.server.Server.suspend",
                     self.proxy.suspend_server,
                     method_args=["value"])

    def test_server_resume(self):
        self._verify("openstack.compute.v2.server.Server.resume",
                     self.proxy.resume_server,
                     method_args=["value"])

    def test_server_lock(self):
        self._verify("openstack.compute.v2.server.Server.lock",
                     self.proxy.lock_server,
                     method_args=["value"])

    def test_server_unlock(self):
        self._verify("openstack.compute.v2.server.Server.unlock",
                     self.proxy.unlock_server,
                     method_args=["value"])

    def test_server_rescue(self):
        self._verify("openstack.compute.v2.server.Server.rescue",
                     self.proxy.rescue_server,
                     method_args=["value"],
                     expected_kwargs={"admin_pass": None, "image_ref": None})

    def test_server_rescue_with_options(self):
        self._verify("openstack.compute.v2.server.Server.rescue",
                     self.proxy.rescue_server,
                     method_args=["value", 'PASS', 'IMG'],
                     expected_kwargs={"admin_pass": 'PASS',
                                      "image_ref": 'IMG'})

    def test_server_unrescue(self):
        self._verify("openstack.compute.v2.server.Server.unrescue",
                     self.proxy.unrescue_server,
                     method_args=["value"])

    def test_server_evacuate(self):
        self._verify("openstack.compute.v2.server.Server.evacuate",
                     self.proxy.evacuate_server,
                     method_args=["value"],
                     expected_kwargs={"host": None, "admin_pass": None,
                                      "force": None})

    def test_server_evacuate_with_options(self):
        self._verify("openstack.compute.v2.server.Server.evacuate",
                     self.proxy.evacuate_server,
                     method_args=["value", 'HOST2', 'NEW_PASS', True],
                     expected_kwargs={"host": "HOST2",
                                      "admin_pass": 'NEW_PASS',
                                      "force": True})

    def test_server_start(self):
        self._verify("openstack.compute.v2.server.Server.start",
                     self.proxy.start_server,
                     method_args=["value"])

    def test_server_stop(self):
        self._verify("openstack.compute.v2.server.Server.stop",
                     self.proxy.stop_server,
                     method_args=["value"])

    def test_server_shelve(self):
        self._verify("openstack.compute.v2.server.Server.shelve",
                     self.proxy.shelve_server,
                     method_args=["value"])

    def test_server_unshelve(self):
        self._verify("openstack.compute.v2.server.Server.unshelve",
                     self.proxy.unshelve_server,
                     method_args=["value"])

    def test_get_server_output(self):
        self._verify("openstack.compute.v2.server.Server.get_console_output",
                     self.proxy.get_server_console_output,
                     method_args=["value"],
                     expected_kwargs={"length": None})

        self._verify("openstack.compute.v2.server.Server.get_console_output",
                     self.proxy.get_server_console_output,
                     method_args=["value", 1],
                     expected_kwargs={"length": 1})

    def test_availability_zones_not_detailed(self):
        self.verify_list(self.proxy.availability_zones,
                         az.AvailabilityZone,
                         method_kwargs={"details": False})

    def test_availability_zones_detailed(self):
        self.verify_list(self.proxy.availability_zones,
                         az.AvailabilityZoneDetail,
                         method_kwargs={"details": True})

    def test_get_all_server_metadata(self):
        self._verify2("openstack.compute.v2.server.Server.get_metadata",
                      self.proxy.get_server_metadata,
                      method_args=["value"],
                      method_result=server.Server(id="value", metadata={}),
                      expected_args=[self.proxy],
                      expected_result={})

    def test_set_server_metadata(self):
        kwargs = {"a": "1", "b": "2"}
        id = "an_id"
        self._verify2("openstack.compute.v2.server.Server.set_metadata",
                      self.proxy.set_server_metadata,
                      method_args=[id],
                      method_kwargs=kwargs,
                      method_result=server.Server.existing(id=id,
                                                           metadata=kwargs),
                      expected_args=[self.proxy],
                      expected_kwargs=kwargs,
                      expected_result=kwargs)

    def test_delete_server_metadata(self):
        self._verify2("openstack.compute.v2.server.Server.delete_metadata",
                      self.proxy.delete_server_metadata,
                      expected_result=None,
                      method_args=["value", "key"],
                      expected_args=[self.proxy, "key"])

    def test_create_image(self):
        metadata = {'k1': 'v1'}
        with mock.patch('openstack.compute.v2.server.Server.create_image') \
                as ci_mock:

                ci_mock.return_value = 'image_id'
                connection_mock = mock.Mock()
                connection_mock.get_image = mock.Mock(return_value='image')
                connection_mock.wait_for_image = mock.Mock()
                self.proxy._connection = connection_mock

                rsp = self.proxy.create_server_image(
                    'server', 'image_name', metadata, wait=True, timeout=1)

                ci_mock.assert_called_with(
                    self.proxy,
                    'image_name',
                    metadata
                )

                self.proxy._connection.get_image.assert_called_with('image_id')
                self.proxy._connection.wait_for_image.assert_called_with(
                    'image',
                    timeout=1)

                self.assertEqual(connection_mock.wait_for_image(), rsp)

    def test_server_group_create(self):
        self.verify_create(self.proxy.create_server_group,
                           server_group.ServerGroup)

    def test_server_group_delete(self):
        self.verify_delete(self.proxy.delete_server_group,
                           server_group.ServerGroup, False)

    def test_server_group_delete_ignore(self):
        self.verify_delete(self.proxy.delete_server_group,
                           server_group.ServerGroup, True)

    def test_server_group_find(self):
        self.verify_find(self.proxy.find_server_group,
                         server_group.ServerGroup)

    def test_server_group_get(self):
        self.verify_get(self.proxy.get_server_group,
                        server_group.ServerGroup)

    def test_server_groups(self):
        self.verify_list(self.proxy.server_groups, server_group.ServerGroup)

    def test_hypervisors_not_detailed(self):
        self.verify_list(self.proxy.hypervisors, hypervisor.Hypervisor,
                         method_kwargs={"details": False})

    def test_hypervisors_detailed(self):
        self.verify_list(self.proxy.hypervisors, hypervisor.HypervisorDetail,
                         method_kwargs={"details": True})

    def test_find_hypervisor(self):
        self.verify_find(self.proxy.find_hypervisor,
                         hypervisor.Hypervisor)

    def test_get_hypervisor(self):
        self.verify_get(self.proxy.get_hypervisor,
                        hypervisor.Hypervisor)

    def test_services(self):
        self.verify_list_no_kwargs(self.proxy.services,
                                   service.Service)

    def test_enable_service(self):
        self._verify('openstack.compute.v2.service.Service.enable',
                     self.proxy.enable_service,
                     method_args=["value", "host1", "nova-compute"],
                     expected_args=["host1", "nova-compute"])

    def test_disable_service(self):
        self._verify('openstack.compute.v2.service.Service.disable',
                     self.proxy.disable_service,
                     method_args=["value", "host1", "nova-compute"],
                     expected_args=["host1", "nova-compute", None])

    def test_force_service_down(self):
        self._verify('openstack.compute.v2.service.Service.force_down',
                     self.proxy.force_service_down,
                     method_args=["value", "host1", "nova-compute"],
                     expected_args=["host1", "nova-compute"])

    def test_live_migrate_server(self):
        self._verify('openstack.compute.v2.server.Server.live_migrate',
                     self.proxy.live_migrate_server,
                     method_args=["value", "host1", False],
                     expected_args=["host1"],
                     expected_kwargs={'force': False, 'block_migration': None})

    def test_fetch_security_groups(self):
        self._verify(
            'openstack.compute.v2.server.Server.fetch_security_groups',
            self.proxy.fetch_server_security_groups,
            method_args=["value"],
            expected_args=[])

    def test_add_security_groups(self):
        self._verify(
            'openstack.compute.v2.server.Server.add_security_group',
            self.proxy.add_security_group_to_server,
            method_args=["value", {'id': 'id', 'name': 'sg'}],
            expected_args=['sg'])

    def test_remove_security_groups(self):
        self._verify(
            'openstack.compute.v2.server.Server.remove_security_group',
            self.proxy.remove_security_group_from_server,
            method_args=["value", {'id': 'id', 'name': 'sg'}],
            expected_args=['sg'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import testtools

from openstack.cloud import exc
from openstack.tests.unit import base


CLUSTERING_DICT = {
    'name': 'fake-name',
    'profile_id': '1',
    'desired_capacity': 1,
    'config': 'fake-config',
    'max_size': 1,
    'min_size': 1,
    'timeout': 100,
    'metadata': {}
}

PROFILE_DICT = {
    'name': 'fake-profile-name',
    'spec': {},
    'metadata': {}
}

POLICY_DICT = {
    'name': 'fake-profile-name',
    'spec': {},
}

RECEIVER_DICT = {
    'action': 'FAKE_CLUSTER_ACTION',
    'cluster_id': 'fake-cluster-id',
    'name': 'fake-receiver-name',
    'params': {},
    'type': 'webhook'
}

NEW_CLUSTERING_DICT = copy.copy(CLUSTERING_DICT)
NEW_CLUSTERING_DICT['id'] = '1'
NEW_PROFILE_DICT = copy.copy(PROFILE_DICT)
NEW_PROFILE_DICT['id'] = '1'
NEW_POLICY_DICT = copy.copy(POLICY_DICT)
NEW_POLICY_DICT['id'] = '1'
NEW_RECEIVER_DICT = copy.copy(RECEIVER_DICT)
NEW_RECEIVER_DICT['id'] = '1'


class TestClustering(base.TestCase):

    def assertAreInstances(self, elements, elem_type):
        for e in elements:
            self.assertIsInstance(e, elem_type)

    def setUp(self):
        super(TestClustering, self).setUp()
        self.use_senlin()

    def test_create_cluster(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles', '1']),
                 json={
                     "profiles": [NEW_PROFILE_DICT]}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles']),
                 json={
                     "profiles": [NEW_PROFILE_DICT]}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters']),
                 json=NEW_CLUSTERING_DICT)
        ])
        profile = self.cloud.get_cluster_profile_by_id(NEW_PROFILE_DICT['id'])
        c = self.cloud.create_cluster(
            name=CLUSTERING_DICT['name'],
            desired_capacity=CLUSTERING_DICT['desired_capacity'],
            profile=profile,
            config=CLUSTERING_DICT['config'],
            max_size=CLUSTERING_DICT['max_size'],
            min_size=CLUSTERING_DICT['min_size'],
            metadata=CLUSTERING_DICT['metadata'],
            timeout=CLUSTERING_DICT['timeout'])

        self.assertEqual(NEW_CLUSTERING_DICT, c)
        self.assert_calls()

    def test_create_cluster_exception(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles', '1']),
                 json={
                     "profiles": [NEW_PROFILE_DICT]}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles']),
                 json={
                     "profiles": [NEW_PROFILE_DICT]}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters']),
                 status_code=500)
        ])
        profile = self.cloud.get_cluster_profile_by_id(NEW_PROFILE_DICT['id'])
        with testtools.ExpectedException(
                exc.OpenStackCloudHTTPError,
                "Error creating cluster fake-name.*"):
            self.cloud.create_cluster(name='fake-name', profile=profile)
        self.assert_calls()

    def test_get_cluster_by_id(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1']),
                 json={
                     "cluster": NEW_CLUSTERING_DICT})
        ])
        cluster = self.cloud.get_cluster_by_id('1')
        self.assertEqual(cluster['id'], '1')
        self.assert_calls()

    def test_get_cluster_not_found_returns_false(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters',
                                                     'no-cluster']),
                 status_code=404)
        ])
        c = self.cloud.get_cluster_by_id('no-cluster')
        self.assertFalse(c)
        self.assert_calls()

    def test_update_cluster(self):
        new_max_size = 5
        updated_cluster = copy.copy(NEW_CLUSTERING_DICT)
        updated_cluster['max_size'] = new_max_size
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1']),
                 json={
                     "cluster": NEW_CLUSTERING_DICT}),
            dict(method='PATCH',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1']),
                 json=updated_cluster,
                 )
        ])
        cluster = self.cloud.get_cluster_by_id('1')
        c = self.cloud.update_cluster(cluster, new_max_size)
        self.assertEqual(updated_cluster, c)
        self.assert_calls()

    def test_delete_cluster(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters']),
                 json={
                     "clusters": [NEW_CLUSTERING_DICT]}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1',
                                                     'policies']),
                 json={"cluster_policies": []}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers']),
                 json={"receivers": []}),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1']),
                 json=NEW_CLUSTERING_DICT)
        ])
        self.assertTrue(self.cloud.delete_cluster('1'))
        self.assert_calls()

    def test_list_clusters(self):
        clusters = {'clusters': [NEW_CLUSTERING_DICT]}
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters']),
                 json=clusters)
        ])
        c = self.cloud.list_clusters()

        self.assertIsInstance(c, list)
        self.assertAreInstances(c, dict)

        self.assert_calls()

    def test_attach_policy_to_cluster(self):
        policy = {
            'policy_id': '1',
            'enabled': 'true'
        }
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1']),
                 json={
                     "cluster": NEW_CLUSTERING_DICT}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies', '1']),
                 json={
                     "policy": NEW_POLICY_DICT}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1',
                                                     'actions']),
                 json={'policy_attach': policy})
        ])
        cluster = self.cloud.get_cluster_by_id('1')
        policy = self.cloud.get_cluster_policy_by_id('1')
        p = self.cloud.attach_policy_to_cluster(cluster, policy, 'true')
        self.assertTrue(p)
        self.assert_calls()

    def test_detach_policy_from_cluster(self):
        updated_cluster = copy.copy(NEW_CLUSTERING_DICT)
        updated_cluster['policies'] = ['1']
        detached_cluster = copy.copy(NEW_CLUSTERING_DICT)
        detached_cluster['policies'] = []

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1']),
                 json={
                     "cluster": NEW_CLUSTERING_DICT}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies', '1']),
                 json={
                     "policy": NEW_POLICY_DICT}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1',
                                                     'actions']),
                 json={'policy_detach': {'policy_id': '1'}}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1']),
                 json={
                     "cluster": updated_cluster}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1']),
                 json={
                     "cluster": detached_cluster}),
        ])
        cluster = self.cloud.get_cluster_by_id('1')
        policy = self.cloud.get_cluster_policy_by_id('1')
        p = self.cloud.detach_policy_from_cluster(cluster, policy, wait=True)
        self.assertTrue(p)
        self.assert_calls()

    def test_get_policy_on_cluster_by_id(self):
        cluster_policy = {
            "cluster_id": "1",
            "cluster_name": "cluster1",
            "enabled": True,
            "id": "1",
            "policy_id": "1",
            "policy_name": "policy1",
            "policy_type": "senlin.policy.deletion-1.0"
        }

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1',
                                                     'policies', '1']),
                 json={
                     "cluster_policy": cluster_policy})
        ])
        policy = self.cloud.get_policy_on_cluster('1', '1')
        self.assertEqual(policy['cluster_id'], '1')
        self.assert_calls()

    def test_get_policy_on_cluster_not_found_returns_false(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1',
                                                     'policies', 'no-policy']),
                 status_code=404)
        ])
        p = self.cloud.get_policy_on_cluster('1', 'no-policy')
        self.assertFalse(p)
        self.assert_calls()

    def test_update_policy_on_cluster(self):
        policy = {
            'policy_id': '1',
            'enabled': 'true'
        }
        updated_cluster = copy.copy(NEW_CLUSTERING_DICT)
        updated_cluster['policies'] = policy
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1']),
                 json={
                     "cluster": NEW_CLUSTERING_DICT}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies',
                                                     '1']),
                 json={
                     "policy": NEW_POLICY_DICT}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1',
                                                     'actions']),
                 json={'policies': []})
        ])
        cluster = self.cloud.get_cluster_by_id('1')
        policy = self.cloud.get_cluster_policy_by_id('1')
        p = self.cloud.update_policy_on_cluster(cluster, policy, True)
        self.assertTrue(p)
        self.assert_calls()

    def test_get_policy_on_cluster(self):
        cluster_policy = {
            'cluster_id': '1',
            'cluster_name': 'cluster1',
            'enabled': 'true',
            'id': '1',
            'policy_id': '1',
            'policy_name': 'policy1',
            'policy_type': 'type'
        }

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters', '1',
                                                     'policies', '1']),
                 json={
                     "cluster_policy": cluster_policy})
        ])
        get_policy = self.cloud.get_policy_on_cluster('1', '1')
        self.assertEqual(get_policy, cluster_policy)
        self.assert_calls()

    def test_create_cluster_profile(self):
        self.register_uris([
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles']),
                 json={'profile': NEW_PROFILE_DICT})
        ])
        p = self.cloud.create_cluster_profile('fake-profile-name', {})

        self.assertEqual(NEW_PROFILE_DICT, p)
        self.assert_calls()

    def test_create_cluster_profile_exception(self):
        self.register_uris([
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles']),
                 status_code=500)
        ])
        with testtools.ExpectedException(
                exc.OpenStackCloudHTTPError,
                "Error creating profile fake-profile-name.*"):
            self.cloud.create_cluster_profile('fake-profile-name', {})
        self.assert_calls()

    def test_list_cluster_profiles(self):
        profiles = {'profiles': [NEW_PROFILE_DICT]}
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles']),
                 json=profiles)
        ])
        p = self.cloud.list_cluster_profiles()

        self.assertIsInstance(p, list)
        self.assertAreInstances(p, dict)

        self.assert_calls()

    def test_get_cluster_profile_by_id(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles', '1']),
                 json={
                     "profile": NEW_PROFILE_DICT})
        ])
        p = self.cloud.get_cluster_profile_by_id('1')
        self.assertEqual(p['id'], '1')
        self.assert_calls()

    def test_get_cluster_profile_not_found_returns_false(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles',
                                                     'no-profile']),
                 status_code=404)
        ])
        p = self.cloud.get_cluster_profile_by_id('no-profile')
        self.assertFalse(p)
        self.assert_calls()

    def test_update_cluster_profile(self):
        new_name = "new-name"
        updated_profile = copy.copy(NEW_PROFILE_DICT)
        updated_profile['name'] = new_name
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles']),
                 json={
                     "profiles": [NEW_PROFILE_DICT]}),
            dict(method='PATCH',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles', '1']),
                 json=updated_profile,
                 )
        ])
        p = self.cloud.update_cluster_profile('1', new_name=new_name)
        self.assertEqual(updated_profile, p)
        self.assert_calls()

    def test_delete_cluster_profile(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles', '1']),
                 json={
                     "profile": NEW_PROFILE_DICT}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters']),
                 json={}),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'profiles', '1']),
                 json=NEW_PROFILE_DICT)
        ])
        profile = self.cloud.get_cluster_profile_by_id('1')
        self.assertTrue(self.cloud.delete_cluster_profile(profile))
        self.assert_calls()

    def test_create_cluster_policy(self):
        self.register_uris([
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies']),
                 json={'policy': NEW_POLICY_DICT})
        ])
        p = self.cloud.create_cluster_policy('fake-policy-name', {})

        self.assertEqual(NEW_POLICY_DICT, p)
        self.assert_calls()

    def test_create_cluster_policy_exception(self):
        self.register_uris([
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies']),
                 status_code=500)
        ])
        with testtools.ExpectedException(
                exc.OpenStackCloudHTTPError,
                "Error creating policy fake-policy-name.*"):
            self.cloud.create_cluster_policy('fake-policy-name', {})
        self.assert_calls()

    def test_list_cluster_policies(self):
        policies = {'policies': [NEW_POLICY_DICT]}
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies']),
                 json=policies)
        ])
        p = self.cloud.list_cluster_policies()

        self.assertIsInstance(p, list)
        self.assertAreInstances(p, dict)

        self.assert_calls()

    def test_get_cluster_policy_by_id(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies', '1']),
                 json={
                     "policy": NEW_POLICY_DICT})
        ])
        p = self.cloud.get_cluster_policy_by_id('1')
        self.assertEqual(p['id'], '1')
        self.assert_calls()

    def test_get_cluster_policy_not_found_returns_false(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies',
                                                     'no-policy']),
                 status_code=404)
        ])
        p = self.cloud.get_cluster_policy_by_id('no-policy')
        self.assertFalse(p)
        self.assert_calls()

    def test_update_cluster_policy(self):
        new_name = "new-name"
        updated_policy = copy.copy(NEW_POLICY_DICT)
        updated_policy['name'] = new_name
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies']),
                 json={
                     "policies": [NEW_POLICY_DICT]}),
            dict(method='PATCH',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies', '1']),
                 json=updated_policy,
                 )
        ])
        p = self.cloud.update_cluster_policy('1', new_name=new_name)
        self.assertEqual(updated_policy, p)
        self.assert_calls()

    def test_delete_cluster_policy(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies', '1']),
                 json={
                     "policy": NEW_POLICY_DICT}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters']),
                 json={}),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'policies', '1']),
                 json=NEW_POLICY_DICT)
        ])
        self.assertTrue(self.cloud.delete_cluster_policy('1'))
        self.assert_calls()

    def test_create_cluster_receiver(self):
        clusters = {'clusters': [NEW_CLUSTERING_DICT]}
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters']),
                 json=clusters),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers']),
                 json={'receiver': NEW_RECEIVER_DICT})
        ])
        r = self.cloud.create_cluster_receiver('fake-receiver-name', {})

        self.assertEqual(NEW_RECEIVER_DICT, r)
        self.assert_calls()

    def test_create_cluster_receiver_exception(self):
        clusters = {'clusters': [NEW_CLUSTERING_DICT]}
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'clusters']),
                 json=clusters),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers']),
                 status_code=500),
        ])
        with testtools.ExpectedException(
                exc.OpenStackCloudHTTPError,
                "Error creating receiver fake-receiver-name.*"):
            self.cloud.create_cluster_receiver('fake-receiver-name', {})
        self.assert_calls()

    def test_list_cluster_receivers(self):
        receivers = {'receivers': [NEW_RECEIVER_DICT]}
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers']),
                 json=receivers)
        ])
        r = self.cloud.list_cluster_receivers()

        self.assertIsInstance(r, list)
        self.assertAreInstances(r, dict)

        self.assert_calls()

    def test_get_cluster_receiver_by_id(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers', '1']),
                 json={
                     "receiver": NEW_RECEIVER_DICT})
        ])
        r = self.cloud.get_cluster_receiver_by_id('1')
        self.assertEqual(r['id'], '1')
        self.assert_calls()

    def test_get_cluster_receiver_not_found_returns_false(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers',
                                                     'no-receiver']),
                 json={'receivers': []})
        ])
        p = self.cloud.get_cluster_receiver_by_id('no-receiver')
        self.assertFalse(p)
        self.assert_calls()

    def test_update_cluster_receiver(self):
        new_name = "new-name"
        updated_receiver = copy.copy(NEW_RECEIVER_DICT)
        updated_receiver['name'] = new_name
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers']),
                 json={
                     "receivers": [NEW_RECEIVER_DICT]}),
            dict(method='PATCH',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers', '1']),
                 json=updated_receiver,
                 )
        ])
        r = self.cloud.update_cluster_receiver('1', new_name=new_name)
        self.assertEqual(updated_receiver, r)
        self.assert_calls()

    def test_delete_cluster_receiver(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers']),
                 json={
                     "receivers": [NEW_RECEIVER_DICT]}),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers', '1']),
                 json=NEW_RECEIVER_DICT),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'clustering', 'public', append=['v1', 'receivers', '1']),
                 json={}),
        ])
        self.assertTrue(self.cloud.delete_cluster_receiver('1', wait=True))
        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import format

from six.moves.urllib import parse


class HREFToUUID(format.Formatter):

    @classmethod
    def deserialize(cls, value):
        """Convert a HREF to the UUID portion"""
        parts = parse.urlsplit(value)

        # Only try to proceed if we have an actual URI.
        # Just check that we have a scheme, netloc, and path.
        if not all(parts[:3]):
            raise ValueError("Unable to convert %s to an ID" % value)

        # The UUID will be the last portion of the URI.
        return parts.path.split("/")[-1]

    @classmethod
    def serialize(cls, value):
        # NOTE(briancurtin): If we had access to the session to get
        # the endpoint we could do something smart here like take an ID
        # and give back an HREF, but this will just have to be something
        # that works different because Barbican does what it does...
        return value



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack import exceptions
from openstack import resource
from openstack import utils


class Backup(resource.Resource):
    """Volume Backup"""
    resource_key = "backup"
    resources_key = "backups"
    base_path = "/backups"

    _query_mapping = resource.QueryParameters(
        'all_tenants', 'limit', 'marker', 'project_id',
        'name', 'status', 'volume_id',
        'sort_key', 'sort_dir')

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_list = True
    allow_get = True

    #: Properties
    #: backup availability zone
    availability_zone = resource.Body("availability_zone")
    #: The container backup in
    container = resource.Body("container")
    #: The date and time when the resource was created.
    created_at = resource.Body("created_at")
    #: data timestamp
    #: The time when the data on the volume was first saved.
    #: If it is a backup from volume, it will be the same as created_at
    #: for a backup. If it is a backup from a snapshot,
    #: it will be the same as created_at for the snapshot.
    data_timestamp = resource.Body('data_timestamp')
    #: backup description
    description = resource.Body("description")
    #: Backup fail reason
    fail_reason = resource.Body("fail_reason")
    #: Force backup
    force = resource.Body("force", type=bool)
    #: has_dependent_backups
    #: If this value is true, there are other backups depending on this backup.
    has_dependent_backups = resource.Body('has_dependent_backups', type=bool)
    #: Indicates whether the backup mode is incremental.
    #: If this value is true, the backup mode is incremental.
    #: If this value is false, the backup mode is full.
    is_incremental = resource.Body("is_incremental", type=bool)
    #: A list of links associated with this volume. *Type: list*
    links = resource.Body("links", type=list)
    #: backup name
    name = resource.Body("name")
    #: backup object count
    object_count = resource.Body("object_count", type=int)
    #: The size of the volume, in gibibytes (GiB).
    size = resource.Body("size", type=int)
    #: The UUID of the source volume snapshot.
    snapshot_id = resource.Body("snapshot_id")
    #: backup status
    #: values: creating, available, deleting, error, restoring, error_restoring
    status = resource.Body("status")
    #: The date and time when the resource was updated.
    updated_at = resource.Body("updated_at")
    #: The UUID of the volume.
    volume_id = resource.Body("volume_id")

    def restore(self, session, volume_id=None, name=None):
        """Restore current backup to volume

        :param session: openstack session
        :param volume_id: The ID of the volume to restore the backup to.
        :param name: The name for new volume creation to restore.
        :return: Updated backup instance
        """
        url = utils.urljoin(self.base_path, self.id, "restore")
        body = {'restore': {}}
        if volume_id:
            body['restore']['volume_id'] = volume_id
        if name:
            body['restore']['name'] = name
        if not (volume_id or name):
            raise exceptions.SDKException('Either of `name` or `volume_id`'
                                          ' must be specified.')
        response = session.post(url,
                                json=body)
        self._translate_response(response, has_body=False)
        return self


BackupDetail = Backup



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import types so that we can reference ListType in sphinx param declarations.
# We can't just use list, because sphinx gets confused by
# openstack.resource.Resource.list and openstack.resource2.Resource.list
import jsonpatch
import types  # noqa
import warnings

from openstack.cloud import exc
from openstack.cloud import _normalize
from openstack.cloud import _utils
from openstack import utils


class BaremetalCloudMixin(_normalize.Normalizer):

    @property
    def _baremetal_client(self):
        if 'baremetal' not in self._raw_clients:
            client = self._get_raw_client('baremetal')
            # Do this to force version discovery. We need to do that, because
            # the endpoint-override trick we do for neutron because
            # ironicclient just appends a /v1 won't work and will break
            # keystoneauth - because ironic's versioned discovery endpoint
            # is non-compliant and doesn't return an actual version dict.
            client = self._get_versioned_client(
                'baremetal', min_version=1, max_version='1.latest')
            self._raw_clients['baremetal'] = client
        return self._raw_clients['baremetal']

    def list_nics(self):
        """Return a list of all bare metal ports."""
        return [nic._to_munch() for nic in self.baremetal.ports(details=True)]

    def list_nics_for_machine(self, uuid):
        """Returns a list of ports present on the machine node.

        :param uuid: String representing machine UUID value in
                     order to identify the machine.
        :returns: A list of ports.
        """
        # TODO(dtantsur): support node names here.
        return [nic._to_munch()
                for nic in self.baremetal.ports(details=True, node_id=uuid)]

    def get_nic_by_mac(self, mac):
        """Get bare metal NIC by its hardware address (usually MAC)."""
        results = [nic._to_munch()
                   for nic in self.baremetal.ports(address=mac, details=True)]
        try:
            return results[0]
        except IndexError:
            return None

    def list_machines(self):
        """List Machines.

        :returns: list of ``munch.Munch`` representing machines.
        """
        return [self._normalize_machine(node)
                for node in self.baremetal.nodes()]

    def get_machine(self, name_or_id):
        """Get Machine by name or uuid

        Search the baremetal host out by utilizing the supplied id value
        which can consist of a name or UUID.

        :param name_or_id: A node name or UUID that will be looked up.

        :returns: ``munch.Munch`` representing the node found or None if no
                  nodes are found.
        """
        try:
            return self._normalize_machine(self.baremetal.get_node(name_or_id))
        except exc.OpenStackCloudResourceNotFound:
            return None

    def get_machine_by_mac(self, mac):
        """Get machine by port MAC address

        :param mac: Port MAC address to query in order to return a node.

        :returns: ``munch.Munch`` representing the node found or None
                  if the node is not found.
        """
        nic = self.get_nic_by_mac(mac)
        if nic is None:
            return None
        else:
            return self.get_machine(nic['node_uuid'])

    def inspect_machine(self, name_or_id, wait=False, timeout=3600):
        """Inspect a Barmetal machine

        Engages the Ironic node inspection behavior in order to collect
        metadata about the baremetal machine.

        :param name_or_id: String representing machine name or UUID value in
                           order to identify the machine.

        :param wait: Boolean value controlling if the method is to wait for
                     the desired state to be reached or a failure to occur.

        :param timeout: Integer value, defautling to 3600 seconds, for the$
                        wait state to reach completion.

        :returns: ``munch.Munch`` representing the current state of the machine
                  upon exit of the method.
        """

        return_to_available = False

        node = self.baremetal.get_node(name_or_id)

        # NOTE(TheJulia): If in available state, we can do this. However,
        # we need to to move the machine back to manageable first.
        if node.provision_state == 'available':
            if node.instance_id:
                raise exc.OpenStackCloudException(
                    "Refusing to inspect available machine %(node)s "
                    "which is associated with an instance "
                    "(instance_uuid %(inst)s)" %
                    {'node': node.id, 'inst': node.instance_id})

            return_to_available = True
            # NOTE(TheJulia): Changing available machine to managedable state
            # and due to state transitions we need to until that transition has
            # completed.
            node = self.baremetal.set_node_provision_state(node, 'manage',
                                                           wait=True,
                                                           timeout=timeout)

        if node.provision_state not in ('manageable', 'inspect failed'):
            raise exc.OpenStackCloudException(
                "Machine %(node)s must be in 'manageable', 'inspect failed' "
                "or 'available' provision state to start inspection, the "
                "current state is %(state)s" %
                {'node': node.id, 'state': node.provision_state})

        node = self.baremetal.set_node_provision_state(node, 'inspect',
                                                       wait=True,
                                                       timeout=timeout)

        if return_to_available:
            node = self.baremetal.set_node_provision_state(node, 'provide',
                                                           wait=True,
                                                           timeout=timeout)

        return self._normalize_machine(node)

    def register_machine(self, nics, wait=False, timeout=3600,
                         lock_timeout=600, **kwargs):
        """Register Baremetal with Ironic

        Allows for the registration of Baremetal nodes with Ironic
        and population of pertinant node information or configuration
        to be passed to the Ironic API for the node.

        This method also creates ports for a list of MAC addresses passed
        in to be utilized for boot and potentially network configuration.

        If a failure is detected creating the network ports, any ports
        created are deleted, and the node is removed from Ironic.

        :param nics:
           An array of MAC addresses that represent the
           network interfaces for the node to be created.

           Example::

              [
                  {'mac': 'aa:bb:cc:dd:ee:01'},
                  {'mac': 'aa:bb:cc:dd:ee:02'}
              ]

        :param wait: Boolean value, defaulting to false, to wait for the
                     node to reach the available state where the node can be
                     provisioned. It must be noted, when set to false, the
                     method will still wait for locks to clear before sending
                     the next required command.

        :param timeout: Integer value, defautling to 3600 seconds, for the
                        wait state to reach completion.

        :param lock_timeout: Integer value, defaulting to 600 seconds, for
                             locks to clear.

        :param kwargs: Key value pairs to be passed to the Ironic API,
                       including uuid, name, chassis_uuid, driver_info,
                       parameters.

        :raises: OpenStackCloudException on operation error.

        :returns: Returns a ``munch.Munch`` representing the new
                  baremetal node.
        """

        msg = ("Baremetal machine node failed to be created.")
        port_msg = ("Baremetal machine port failed to be created.")

        url = '/nodes'
        # TODO(TheJulia): At some point we need to figure out how to
        # handle data across when the requestor is defining newer items
        # with the older api.
        machine = self._baremetal_client.post(url,
                                              json=kwargs,
                                              error_message=msg,
                                              microversion="1.6")

        created_nics = []
        try:
            for row in nics:
                payload = {'address': row['mac'],
                           'node_uuid': machine['uuid']}
                nic = self._baremetal_client.post('/ports',
                                                  json=payload,
                                                  error_message=port_msg)
                created_nics.append(nic['uuid'])

        except Exception as e:
            self.log.debug("ironic NIC registration failed", exc_info=True)
            # TODO(mordred) Handle failures here
            try:
                for uuid in created_nics:
                    try:
                        port_url = '/ports/{uuid}'.format(uuid=uuid)
                        # NOTE(TheJulia): Added in hope that it is logged.
                        port_msg = ('Failed to delete port {port} for node '
                                    '{node}').format(port=uuid,
                                                     node=machine['uuid'])
                        self._baremetal_client.delete(port_url,
                                                      error_message=port_msg)
                    except Exception:
                        pass
            finally:
                version = "1.6"
                msg = "Baremetal machine failed to be deleted."
                url = '/nodes/{node_id}'.format(
                    node_id=machine['uuid'])
                self._baremetal_client.delete(url,
                                              error_message=msg,
                                              microversion=version)
            raise exc.OpenStackCloudException(
                "Error registering NICs with the baremetal service: %s"
                % str(e))

        with _utils.shade_exceptions(
                "Error transitioning node to available state"):
            if wait:
                for count in utils.iterate_timeout(
                        timeout,
                        "Timeout waiting for node transition to "
                        "available state"):

                    machine = self.get_machine(machine['uuid'])

                    # Note(TheJulia): Per the Ironic state code, a node
                    # that fails returns to enroll state, which means a failed
                    # node cannot be determined at this point in time.
                    if machine['provision_state'] in ['enroll']:
                        self.node_set_provision_state(
                            machine['uuid'], 'manage')
                    elif machine['provision_state'] in ['manageable']:
                        self.node_set_provision_state(
                            machine['uuid'], 'provide')
                    elif machine['last_error'] is not None:
                        raise exc.OpenStackCloudException(
                            "Machine encountered a failure: %s"
                            % machine['last_error'])

                    # Note(TheJulia): Earlier versions of Ironic default to
                    # None and later versions default to available up until
                    # the introduction of enroll state.
                    # Note(TheJulia): The node will transition through
                    # cleaning if it is enabled, and we will wait for
                    # completion.
                    elif machine['provision_state'] in ['available', None]:
                        break

            else:
                if machine['provision_state'] in ['enroll']:
                    self.node_set_provision_state(machine['uuid'], 'manage')
                    # Note(TheJulia): We need to wait for the lock to clear
                    # before we attempt to set the machine into provide state
                    # which allows for the transition to available.
                    for count in utils.iterate_timeout(
                            lock_timeout,
                            "Timeout waiting for reservation to clear "
                            "before setting provide state"):
                        machine = self.get_machine(machine['uuid'])
                        if (machine['reservation'] is None
                                and machine['provision_state'] != 'enroll'):
                            # NOTE(TheJulia): In this case, the node has
                            # has moved on from the previous state and is
                            # likely not being verified, as no lock is
                            # present on the node.
                            self.node_set_provision_state(
                                machine['uuid'], 'provide')
                            machine = self.get_machine(machine['uuid'])
                            break

                        elif machine['provision_state'] in [
                                'cleaning',
                                'available']:
                            break

                        elif machine['last_error'] is not None:
                            raise exc.OpenStackCloudException(
                                "Machine encountered a failure: %s"
                                % machine['last_error'])
        if not isinstance(machine, str):
            return self._normalize_machine(machine)
        else:
            return machine

    def unregister_machine(self, nics, uuid, wait=False, timeout=600):
        """Unregister Baremetal from Ironic

        Removes entries for Network Interfaces and baremetal nodes
        from an Ironic API

        :param nics: An array of strings that consist of MAC addresses
                          to be removed.
        :param string uuid: The UUID of the node to be deleted.

        :param wait: Boolean value, defaults to false, if to block the method
                     upon the final step of unregistering the machine.

        :param timeout: Integer value, representing seconds with a default
                        value of 600, which controls the maximum amount of
                        time to block the method's completion on.

        :raises: OpenStackCloudException on operation failure.
        """

        machine = self.get_machine(uuid)
        invalid_states = ['active', 'cleaning', 'clean wait', 'clean failed']
        if machine['provision_state'] in invalid_states:
            raise exc.OpenStackCloudException(
                "Error unregistering node '%s' due to current provision "
                "state '%s'" % (uuid, machine['provision_state']))

        # NOTE(TheJulia) There is a high possibility of a lock being present
        # if the machine was just moved through the state machine. This was
        # previously concealed by exception retry logic that detected the
        # failure, and resubitted the request in python-ironicclient.
        try:
            self.wait_for_baremetal_node_lock(machine, timeout=timeout)
        except exc.OpenStackCloudException as e:
            raise exc.OpenStackCloudException(
                "Error unregistering node '%s': Exception occured while"
                " waiting to be able to proceed: %s" % (machine['uuid'], e))

        for nic in nics:
            port_msg = ("Error removing NIC {nic} from baremetal API for "
                        "node {uuid}").format(nic=nic, uuid=uuid)
            port_url = '/ports/detail?address={mac}'.format(mac=nic['mac'])
            port = self._baremetal_client.get(port_url, microversion=1.6,
                                              error_message=port_msg)
            port_url = '/ports/{uuid}'.format(uuid=port['ports'][0]['uuid'])
            _utils._call_client_and_retry(self._baremetal_client.delete,
                                          port_url, retry_on=[409, 503],
                                          error_message=port_msg)

        with _utils.shade_exceptions(
                "Error unregistering machine {node_id} from the baremetal "
                "API".format(node_id=uuid)):

            # NOTE(TheJulia): While this should not matter microversion wise,
            # ironic assumes all calls without an explicit microversion to be
            # version 1.0. Ironic expects to deprecate support for older
            # microversions in future releases, as such, we explicitly set
            # the version to what we have been using with the client library..
            version = "1.6"
            msg = "Baremetal machine failed to be deleted"
            url = '/nodes/{node_id}'.format(
                node_id=uuid)
            _utils._call_client_and_retry(self._baremetal_client.delete,
                                          url, retry_on=[409, 503],
                                          error_message=msg,
                                          microversion=version)

            if wait:
                for count in utils.iterate_timeout(
                        timeout,
                        "Timeout waiting for machine to be deleted"):
                    if not self.get_machine(uuid):
                        break

    def patch_machine(self, name_or_id, patch):
        """Patch Machine Information

        This method allows for an interface to manipulate node entries
        within Ironic.

        :param string name_or_id: A machine name or UUID to be updated.
        :param patch:
           The JSON Patch document is a list of dictonary objects
           that comply with RFC 6902 which can be found at
           https://tools.ietf.org/html/rfc6902.

           Example patch construction::

               patch=[]
               patch.append({
                   'op': 'remove',
                   'path': '/instance_info'
               })
               patch.append({
                   'op': 'replace',
                   'path': '/name',
                   'value': 'newname'
               })
               patch.append({
                   'op': 'add',
                   'path': '/driver_info/username',
                   'value': 'administrator'
               })

        :raises: OpenStackCloudException on operation error.

        :returns: ``munch.Munch`` representing the newly updated node.
        """
        return self._normalize_machine(
            self.baremetal.patch_node(name_or_id, patch))

    def update_machine(self, name_or_id, **attrs):
        """Update a machine with new configuration information

        A user-friendly method to perform updates of a machine, in whole or
        part.

        :param string name_or_id: A machine name or UUID to be updated.
        :param attrs: Attributes to updated on the machine.

        :raises: OpenStackCloudException on operation error.

        :returns: ``munch.Munch`` containing a machine sub-dictonary consisting
                  of the updated data returned from the API update operation,
                  and a list named changes which contains all of the API paths
                  that received updates.
        """
        machine = self.get_machine(name_or_id)
        if not machine:
            raise exc.OpenStackCloudException(
                "Machine update failed to find Machine: %s. " % name_or_id)

        new_config = dict(machine, **attrs)

        try:
            patch = jsonpatch.JsonPatch.from_diff(machine, new_config)
        except Exception as e:
            raise exc.OpenStackCloudException(
                "Machine update failed - Error generating JSON patch object "
                "for submission to the API. Machine: %s Error: %s"
                % (name_or_id, e))

        if not patch:
            return dict(
                node=machine,
                changes=None
            )

        change_list = [change['path'] for change in patch]
        node = self.baremetal.update_node(machine, **attrs)
        return dict(
            node=self._normalize_machine(node),
            changes=change_list
        )

    def attach_port_to_machine(self, name_or_id, port_name_or_id):
        """Attach a virtual port to the bare metal machine.

        :param string name_or_id: A machine name or UUID.
        :param string port_name_or_id: A port name or UUID.
            Note that this is a Network service port, not a bare metal NIC.
        :return: Nothing.
        """
        machine = self.get_machine(name_or_id)
        port = self.get_port(port_name_or_id)
        self.baremetal.attach_vif_to_node(machine, port['id'])

    def detach_port_from_machine(self, name_or_id, port_name_or_id):
        """Detach a virtual port from the bare metal machine.

        :param string name_or_id: A machine name or UUID.
        :param string port_name_or_id: A port name or UUID.
            Note that this is a Network service port, not a bare metal NIC.
        :return: Nothing.
        """
        machine = self.get_machine(name_or_id)
        port = self.get_port(port_name_or_id)
        self.baremetal.detach_vif_from_node(machine, port['id'])

    def list_ports_attached_to_machine(self, name_or_id):
        """List virtual ports attached to the bare metal machine.

        :param string name_or_id: A machine name or UUID.
        :returns: List of ``munch.Munch`` representing the ports.
        """
        machine = self.get_machine(name_or_id)
        vif_ids = self.baremetal.list_node_vifs(machine)
        return [self.get_port(vif) for vif in vif_ids]

    def validate_machine(self, name_or_id, for_deploy=True):
        """Validate parameters of the machine.

        :param string name_or_id: The Name or UUID value representing the
                                  baremetal node.
        :param bool for_deploy: If ``True``, validate readiness for deployment,
                                otherwise validate only the power management
                                properties.
        :raises: :exc:`~openstack.exceptions.ValidationException`
        """
        if for_deploy:
            ifaces = ('boot', 'deploy', 'management', 'power')
        else:
            ifaces = ('power',)
        self.baremetal.validate_node(name_or_id, required=ifaces)

    def validate_node(self, uuid):
        warnings.warn('validate_node is deprecated, please use '
                      'validate_machine instead', DeprecationWarning)
        self.baremetal.validate_node(uuid)

    def node_set_provision_state(self,
                                 name_or_id,
                                 state,
                                 configdrive=None,
                                 wait=False,
                                 timeout=3600):
        """Set Node Provision State

        Enables a user to provision a Machine and optionally define a
        config drive to be utilized.

        :param string name_or_id: The Name or UUID value representing the
                                  baremetal node.
        :param string state: The desired provision state for the
                             baremetal node.
        :param string configdrive: An optional URL or file or path
                                   representing the configdrive. In the
                                   case of a directory, the client API
                                   will create a properly formatted
                                   configuration drive file and post the
                                   file contents to the API for
                                   deployment.
        :param boolean wait: A boolean value, defaulted to false, to control
                             if the method will wait for the desire end state
                             to be reached before returning.
        :param integer timeout: Integer value, defaulting to 3600 seconds,
                                representing the amount of time to wait for
                                the desire end state to be reached.

        :raises: OpenStackCloudException on operation error.

        :returns: ``munch.Munch`` representing the current state of the machine
                  upon exit of the method.
        """
        node = self.baremetal.set_node_provision_state(
            name_or_id, target=state, config_drive=configdrive,
            wait=wait, timeout=timeout)
        return self._normalize_machine(node)

    def set_machine_maintenance_state(
            self,
            name_or_id,
            state=True,
            reason=None):
        """Set Baremetal Machine Maintenance State

        Sets Baremetal maintenance state and maintenance reason.

        :param string name_or_id: The Name or UUID value representing the
                                  baremetal node.
        :param boolean state: The desired state of the node.  True being in
                              maintenance where as False means the machine
                              is not in maintenance mode.  This value
                              defaults to True if not explicitly set.
        :param string reason: An optional freeform string that is supplied to
                              the baremetal API to allow for notation as to why
                              the node is in maintenance state.

        :raises: OpenStackCloudException on operation error.

        :returns: None
        """
        if state:
            self.baremetal.set_node_maintenance(name_or_id, reason)
        else:
            self.baremetal.unset_node_maintenance(name_or_id)

    def remove_machine_from_maintenance(self, name_or_id):
        """Remove Baremetal Machine from Maintenance State

        Similarly to set_machine_maintenance_state, this method
        removes a machine from maintenance state.  It must be noted
        that this method simpily calls set_machine_maintenace_state
        for the name_or_id requested and sets the state to False.

        :param string name_or_id: The Name or UUID value representing the
                                  baremetal node.

        :raises: OpenStackCloudException on operation error.

        :returns: None
        """
        self.baremetal.unset_node_maintenance(name_or_id)

    def set_machine_power_on(self, name_or_id):
        """Activate baremetal machine power

        This is a method that sets the node power state to "on".

        :params string name_or_id: A string representing the baremetal
                                   node to have power turned to an "on"
                                   state.

        :raises: OpenStackCloudException on operation error.

        :returns: None
        """
        self.baremetal.set_node_power_state(name_or_id, 'power on')

    def set_machine_power_off(self, name_or_id):
        """De-activate baremetal machine power

        This is a method that sets the node power state to "off".

        :params string name_or_id: A string representing the baremetal
                                   node to have power turned to an "off"
                                   state.

        :raises: OpenStackCloudException on operation error.

        :returns:
        """
        self.baremetal.set_node_power_state(name_or_id, 'power off')

    def set_machine_power_reboot(self, name_or_id):
        """De-activate baremetal machine power

        This is a method that sets the node power state to "reboot", which
        in essence changes the machine power state to "off", and that back
        to "on".

        :params string name_or_id: A string representing the baremetal
                                   node to have power turned to an "off"
                                   state.

        :raises: OpenStackCloudException on operation error.

        :returns: None
        """
        self.baremetal.set_node_power_state(name_or_id, 'rebooting')

    def activate_node(self, uuid, configdrive=None,
                      wait=False, timeout=1200):
        self.node_set_provision_state(
            uuid, 'active', configdrive, wait=wait, timeout=timeout)

    def deactivate_node(self, uuid, wait=False,
                        timeout=1200):
        self.node_set_provision_state(
            uuid, 'deleted', wait=wait, timeout=timeout)

    def set_node_instance_info(self, uuid, patch):
        warnings.warn("The set_node_instance_info call is deprecated, "
                      "use patch_machine or update_machine instead",
                      DeprecationWarning)
        return self.patch_machine(uuid, patch)

    def purge_node_instance_info(self, uuid):
        warnings.warn("The purge_node_instance_info call is deprecated, "
                      "use patch_machine or update_machine instead",
                      DeprecationWarning)
        return self.patch_machine(uuid,
                                  dict(path='/instance_info', op='remove'))

    def wait_for_baremetal_node_lock(self, node, timeout=30):
        """Wait for a baremetal node to have no lock.

        DEPRECATED, use ``wait_for_node_reservation`` on the `baremetal` proxy.

        :raises: OpenStackCloudException upon client failure.
        :returns: None
        """
        warnings.warn("The wait_for_baremetal_node_lock call is deprecated "
                      "in favor of wait_for_node_reservation on the baremetal "
                      "proxy", DeprecationWarning)
        self.baremetal.wait_for_node_reservation(node, timeout)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack import resource


class Subscription(resource.Resource):
    # FIXME(anyone): The name string of `location` field of Zaqar API response
    # is lower case. That is inconsistent with the guide from API-WG. This is
    # a workaround for this issue.
    location = resource.Header("location")

    resources_key = 'subscriptions'
    base_path = '/queues/%(queue_name)s/subscriptions'

    # capabilities
    allow_create = True
    allow_list = True
    allow_fetch = True
    allow_delete = True

    # Properties
    #: The value in seconds indicating how long the subscription has existed.
    age = resource.Body("age")
    #: Alternate id of the subscription. This key is used in response of
    #: subscription create API to return id of subscription created.
    subscription_id = resource.Body("subscription_id", alternate_id=True)
    #: The extra metadata for the subscription. The value must be a dict.
    #: If the subscriber is `mailto`. The options can contain `from` and
    #: `subject` to indicate the email's author and title.
    options = resource.Body("options", type=dict)
    #: The queue name which the subscription is registered on.
    source = resource.Body("source")
    #: The destination of the message. Two kinds of subscribers are supported:
    #: http/https and email. The http/https subscriber should start with
    #: `http/https`. The email subscriber should start with `mailto`.
    subscriber = resource.Body("subscriber")
    #: Number of seconds the subscription remains alive? The ttl value must
    #: be great than 60 seconds. The default value is 3600 seconds.
    ttl = resource.Body("ttl")
    #: The queue name which the subscription is registered on.
    queue_name = resource.URI("queue_name")
    #: The ID to identify the client accessing Zaqar API. Must be specified
    #: in header for each API request.
    client_id = resource.Header("Client-ID")
    #: The ID to identify the project. Must be provided when keystone
    #: authentication is not enabled in Zaqar service.
    project_id = resource.Header("X-PROJECT-ID")

    def create(self, session, prepend_key=True, base_path=None):
        request = self._prepare_request(requires_id=False,
                                        prepend_key=prepend_key,
                                        base_path=base_path)
        headers = {
            "Client-ID": self.client_id or str(uuid.uuid4()),
            "X-PROJECT-ID": self.project_id or session.get_project_id()
        }
        request.headers.update(headers)
        response = session.post(request.url,
                                json=request.body, headers=request.headers)

        self._translate_response(response)
        return self

    @classmethod
    def list(cls, session, paginated=True, base_path=None, **params):
        """This method is a generator which yields subscription objects.

        This is almost the copy of list method of resource.Resource class.
        The only difference is the request header now includes `Client-ID`
        and `X-PROJECT-ID` fields which are required by Zaqar v2 API.
        """
        more_data = True

        if base_path is None:
            base_path = cls.base_path

        uri = base_path % params
        headers = {
            "Client-ID": params.get('client_id', None) or str(uuid.uuid4()),
            "X-PROJECT-ID": params.get('project_id', None
                                       ) or session.get_project_id()
        }

        query_params = cls._query_mapping._transpose(params, cls)
        while more_data:
            resp = session.get(uri,
                               headers=headers, params=query_params)
            resp = resp.json()
            resp = resp[cls.resources_key]

            if not resp:
                more_data = False

            yielded = 0
            new_marker = None
            for data in resp:
                value = cls.existing(**data)
                new_marker = value.id
                yielded += 1
                yield value

            if not paginated:
                return
            if "limit" in query_params and yielded < query_params["limit"]:
                return
            query_params["limit"] = yielded
            query_params["marker"] = new_marker

    def fetch(self, session, requires_id=True,
              base_path=None, error_message=None):
        request = self._prepare_request(requires_id=requires_id,
                                        base_path=base_path)
        headers = {
            "Client-ID": self.client_id or str(uuid.uuid4()),
            "X-PROJECT-ID": self.project_id or session.get_project_id()
        }

        request.headers.update(headers)
        response = session.get(request.url,
                               headers=request.headers)
        self._translate_response(response)

        return self

    def delete(self, session):
        request = self._prepare_request()
        headers = {
            "Client-ID": self.client_id or str(uuid.uuid4()),
            "X-PROJECT-ID": self.project_id or session.get_project_id()
        }

        request.headers.update(headers)
        response = session.delete(request.url,
                                  headers=request.headers)

        self._translate_response(response, has_body=False)
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.unit import base

from openstack.baremetal.v1 import port

FAKE = {
    "address": "11:11:11:11:11:11",
    "created_at": "2016-08-18T22:28:49.946416+00:00",
    "extra": {},
    "internal_info": {},
    "links": [
        {
            "href": "http://127.0.0.1:6385/v1/ports/<PORT_ID>",
            "rel": "self"
        },
        {
            "href": "http://127.0.0.1:6385/ports/<PORT_ID>",
            "rel": "bookmark"
        }
    ],
    "local_link_connection": {
        "port_id": "Ethernet3/1",
        "switch_id": "0a:1b:2c:3d:4e:5f",
        "switch_info": "switch1"
    },
    "node_uuid": "6d85703a-565d-469a-96ce-30b6de53079d",
    "portgroup_uuid": "e43c722c-248e-4c6e-8ce8-0d8ff129387a",
    "pxe_enabled": True,
    "updated_at": None,
    "uuid": "d2b30520-907d-46c8-bfee-c5586e6fb3a1"
}


class TestPort(base.TestCase):

    def test_basic(self):
        sot = port.Port()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('ports', sot.resources_key)
        self.assertEqual('/ports', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertEqual('PATCH', sot.commit_method)

    def test_instantiate(self):
        sot = port.PortDetail(**FAKE)
        self.assertEqual(FAKE['uuid'], sot.id)
        self.assertEqual(FAKE['address'], sot.address)
        self.assertEqual(FAKE['created_at'], sot.created_at)
        self.assertEqual(FAKE['extra'], sot.extra)
        self.assertEqual(FAKE['internal_info'], sot.internal_info)
        self.assertEqual(FAKE['links'], sot.links)
        self.assertEqual(FAKE['local_link_connection'],
                         sot.local_link_connection)
        self.assertEqual(FAKE['node_uuid'], sot.node_id)
        self.assertEqual(FAKE['portgroup_uuid'], sot.port_group_id)
        self.assertEqual(FAKE['pxe_enabled'], sot.is_pxe_enabled)
        self.assertEqual(FAKE['updated_at'], sot.updated_at)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from openstack.tests.unit import base

from openstack.block_storage.v3 import volume

FAKE_ID = "6685584b-1eac-4da6-b5c3-555430cf68ff"
IMAGE_METADATA = {
    'container_format': 'bare',
    'min_ram': '64', 'disk_format': u'qcow2',
    'image_name': 'TestVM',
    'image_id': '625d4f2c-cf67-4af3-afb6-c7220f766947',
    'checksum': '64d7c1cd2b6f60c92c14662941cb7913',
    'min_disk': '0', u'size': '13167616'
}

VOLUME = {
    "status": "creating",
    "name": "my_volume",
    "attachments": [],
    "availability_zone": "nova",
    "bootable": "false",
    "created_at": "2015-03-09T12:14:57.233772",
    "description": "something",
    "volume_type": "some_type",
    "snapshot_id": "93c2e2aa-7744-4fd6-a31a-80c4726b08d7",
    "source_volid": None,
    "imageRef": "some_image",
    "metadata": {},
    "volume_image_metadata": IMAGE_METADATA,
    "id": FAKE_ID,
    "size": 10,
    "os-vol-host-attr:host": "127.0.0.1",
    "os-vol-tenant-attr:tenant_id": "some tenant",
    "os-vol-mig-status-attr:migstat": "done",
    "os-vol-mig-status-attr:name_id": "93c2e2aa-7744-4fd6-a31a-80c4726b08d7",
    "replication_status": "nah",
    "os-volume-replication:extended_status": "really nah",
    "consistencygroup_id": "123asf-asdf123",
    "os-volume-replication:driver_data": "ahasadfasdfasdfasdfsdf",
    "snapshot_id": "93c2e2aa-7744-4fd6-a31a-80c4726b08d7",
    "encrypted": "false"
}


class TestVolume(base.TestCase):

    def setUp(self):
        super(TestVolume, self).setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.sess = mock.Mock()
        self.sess.post = mock.Mock(return_value=self.resp)

    def test_basic(self):
        sot = volume.Volume(VOLUME)
        self.assertEqual("volume", sot.resource_key)
        self.assertEqual("volumes", sot.resources_key)
        self.assertEqual("/volumes", sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual({"name": "name",
                              "status": "status",
                              "all_projects": "all_tenants",
                              "project_id": "project_id",
                              "limit": "limit",
                              "marker": "marker"},
                             sot._query_mapping._mapping)

    def test_create(self):
        sot = volume.Volume(**VOLUME)
        self.assertEqual(VOLUME["id"], sot.id)
        self.assertEqual(VOLUME["status"], sot.status)
        self.assertEqual(VOLUME["attachments"], sot.attachments)
        self.assertEqual(VOLUME["availability_zone"], sot.availability_zone)
        self.assertFalse(sot.is_bootable)
        self.assertEqual(VOLUME["created_at"], sot.created_at)
        self.assertEqual(VOLUME["description"], sot.description)
        self.assertEqual(VOLUME["volume_type"], sot.volume_type)
        self.assertEqual(VOLUME["snapshot_id"], sot.snapshot_id)
        self.assertEqual(VOLUME["source_volid"], sot.source_volume_id)
        self.assertEqual(VOLUME["metadata"], sot.metadata)
        self.assertEqual(VOLUME["volume_image_metadata"],
                         sot.volume_image_metadata)
        self.assertEqual(VOLUME["size"], sot.size)
        self.assertEqual(VOLUME["imageRef"], sot.image_id)
        self.assertEqual(VOLUME["os-vol-host-attr:host"], sot.host)
        self.assertEqual(VOLUME["os-vol-tenant-attr:tenant_id"],
                         sot.project_id)
        self.assertEqual(VOLUME["os-vol-mig-status-attr:migstat"],
                         sot.migration_status)
        self.assertEqual(VOLUME["os-vol-mig-status-attr:name_id"],
                         sot.migration_id)
        self.assertEqual(VOLUME["replication_status"],
                         sot.replication_status)
        self.assertEqual(
            VOLUME["os-volume-replication:extended_status"],
            sot.extended_replication_status)
        self.assertEqual(VOLUME["consistencygroup_id"],
                         sot.consistency_group_id)
        self.assertEqual(VOLUME["os-volume-replication:driver_data"],
                         sot.replication_driver_data)
        self.assertFalse(sot.is_encrypted)

    def test_extend(self):
        sot = volume.Volume(**VOLUME)

        self.assertIsNone(sot.extend(self.sess, '20'))

        url = 'volumes/%s/action' % FAKE_ID
        body = {"os-extend": {"new_size": "20"}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(url, json=body, headers=headers)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random
import uuid

from openstack import exceptions
from openstack.tests.functional.baremetal import base


class TestBareMetalNode(base.BaseBaremetalTest):
    def test_node_create_get_delete(self):
        node = self.create_node(name='node-name')
        self.assertEqual(node.name, 'node-name')
        self.assertEqual(node.driver, 'fake-hardware')
        self.assertEqual(node.provision_state, 'available')
        self.assertFalse(node.is_maintenance)

        # NOTE(dtantsur): get_node and find_node only differ in handing missing
        # nodes, otherwise they are identical.
        for call, ident in [(self.conn.baremetal.get_node, self.node_id),
                            (self.conn.baremetal.get_node, 'node-name'),
                            (self.conn.baremetal.find_node, self.node_id),
                            (self.conn.baremetal.find_node, 'node-name')]:
            found = call(ident)
            self.assertEqual(node.id, found.id)
            self.assertEqual(node.name, found.name)

        with_fields = self.conn.baremetal.get_node(
            'node-name',
            fields=['uuid', 'driver', 'instance_id'])
        self.assertEqual(node.id, with_fields.id)
        self.assertEqual(node.driver, with_fields.driver)
        self.assertIsNone(with_fields.name)
        self.assertIsNone(with_fields.provision_state)

        nodes = self.conn.baremetal.nodes()
        self.assertIn(node.id, [n.id for n in nodes])

        self.conn.baremetal.delete_node(node, ignore_missing=False)
        self.assertRaises(exceptions.ResourceNotFound,
                          self.conn.baremetal.get_node, self.node_id)

    def test_node_update(self):
        node = self.create_node(name='node-name', extra={'foo': 'bar'})
        node.name = 'new-name'
        node.extra = {'answer': 42}
        instance_uuid = str(uuid.uuid4())

        node = self.conn.baremetal.update_node(node,
                                               instance_id=instance_uuid)
        self.assertEqual('new-name', node.name)
        self.assertEqual({'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.conn.baremetal.get_node('new-name')
        self.assertEqual('new-name', node.name)
        self.assertEqual({'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.conn.baremetal.update_node(node,
                                               instance_id=None)
        self.assertIsNone(node.instance_id)

        node = self.conn.baremetal.get_node('new-name')
        self.assertIsNone(node.instance_id)

    def test_node_update_by_name(self):
        self.create_node(name='node-name', extra={'foo': 'bar'})
        instance_uuid = str(uuid.uuid4())

        node = self.conn.baremetal.update_node('node-name',
                                               instance_id=instance_uuid,
                                               extra={'answer': 42})
        self.assertEqual({'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.conn.baremetal.get_node('node-name')
        self.assertEqual({'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.conn.baremetal.update_node('node-name',
                                               instance_id=None)
        self.assertIsNone(node.instance_id)

        node = self.conn.baremetal.get_node('node-name')
        self.assertIsNone(node.instance_id)

    def test_node_patch(self):
        node = self.create_node(name='node-name', extra={'foo': 'bar'})
        node.name = 'new-name'
        instance_uuid = str(uuid.uuid4())

        node = self.conn.baremetal.patch_node(
            node,
            [dict(path='/instance_id', op='replace', value=instance_uuid),
             dict(path='/extra/answer', op='add', value=42)])
        self.assertEqual('new-name', node.name)
        self.assertEqual({'foo': 'bar', 'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.conn.baremetal.get_node('new-name')
        self.assertEqual('new-name', node.name)
        self.assertEqual({'foo': 'bar', 'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.conn.baremetal.patch_node(
            node,
            [dict(path='/instance_id', op='remove'),
             dict(path='/extra/answer', op='remove')])
        self.assertIsNone(node.instance_id)
        self.assertNotIn('answer', node.extra)

        node = self.conn.baremetal.get_node('new-name')
        self.assertIsNone(node.instance_id)
        self.assertNotIn('answer', node.extra)

    def test_node_list_update_delete(self):
        self.create_node(name='node-name', extra={'foo': 'bar'})
        node = next(n for n in
                    self.conn.baremetal.nodes(details=True,
                                              provision_state='available',
                                              is_maintenance=False,
                                              associated=False)
                    if n.name == 'node-name')
        self.assertEqual(node.extra, {'foo': 'bar'})

        # This test checks that resources returned from listing are usable
        self.conn.baremetal.update_node(node, extra={'foo': 42})
        self.conn.baremetal.delete_node(node, ignore_missing=False)

    def test_node_create_in_enroll_provide(self):
        node = self.create_node(provision_state='enroll')
        self.node_id = node.id

        self.assertEqual(node.driver, 'fake-hardware')
        self.assertEqual(node.provision_state, 'enroll')
        self.assertIsNone(node.power_state)
        self.assertFalse(node.is_maintenance)

        self.conn.baremetal.set_node_provision_state(node, 'manage',
                                                     wait=True)
        self.assertEqual(node.provision_state, 'manageable')

        self.conn.baremetal.set_node_provision_state(node, 'provide',
                                                     wait=True)
        self.assertEqual(node.provision_state, 'available')

    def test_node_create_in_enroll_provide_by_name(self):
        name = 'node-%d' % random.randint(0, 1000)
        node = self.create_node(provision_state='enroll', name=name)
        self.node_id = node.id

        self.assertEqual(node.driver, 'fake-hardware')
        self.assertEqual(node.provision_state, 'enroll')
        self.assertIsNone(node.power_state)
        self.assertFalse(node.is_maintenance)

        node = self.conn.baremetal.set_node_provision_state(name, 'manage',
                                                            wait=True)
        self.assertEqual(node.provision_state, 'manageable')

        node = self.conn.baremetal.set_node_provision_state(name, 'provide',
                                                            wait=True)
        self.assertEqual(node.provision_state, 'available')

    def test_node_power_state(self):
        node = self.create_node()
        self.assertIsNone(node.power_state)

        self.conn.baremetal.set_node_power_state(node, 'power on')
        node = self.conn.baremetal.get_node(node.id)
        # Fake nodes react immediately to power requests.
        self.assertEqual('power on', node.power_state)

        self.conn.baremetal.set_node_power_state(node, 'power off')
        node = self.conn.baremetal.get_node(node.id)
        self.assertEqual('power off', node.power_state)

    def test_node_validate(self):
        node = self.create_node()
        # Fake hardware passes validation for all interfaces
        result = self.conn.baremetal.validate_node(node)
        for iface in ('boot', 'deploy', 'management', 'power'):
            self.assertTrue(result[iface].result)
            self.assertFalse(result[iface].reason)

    def test_node_negative_non_existing(self):
        uuid = "5c9dcd04-2073-49bc-9618-99ae634d8971"
        self.assertRaises(exceptions.ResourceNotFound,
                          self.conn.baremetal.get_node, uuid)
        self.assertRaises(exceptions.ResourceNotFound,
                          self.conn.baremetal.find_node, uuid,
                          ignore_missing=False)
        self.assertRaises(exceptions.ResourceNotFound,
                          self.conn.baremetal.delete_node, uuid,
                          ignore_missing=False)
        self.assertRaises(exceptions.ResourceNotFound,
                          self.conn.baremetal.update_node, uuid,
                          name='new-name')
        self.assertIsNone(self.conn.baremetal.find_node(uuid))
        self.assertIsNone(self.conn.baremetal.delete_node(uuid))

    def test_maintenance(self):
        reason = "Prepating for taking over the world"

        node = self.create_node()
        self.assertFalse(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Initial setting without the reason
        node = self.conn.baremetal.set_node_maintenance(node)
        self.assertTrue(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Updating the reason later
        node = self.conn.baremetal.set_node_maintenance(node, reason)
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)

        # Removing the reason later
        node = self.conn.baremetal.set_node_maintenance(node)
        self.assertTrue(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Unsetting maintenance
        node = self.conn.baremetal.unset_node_maintenance(node)
        self.assertFalse(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Initial setting with the reason
        node = self.conn.baremetal.set_node_maintenance(node, reason)
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)

    def test_maintenance_via_update(self):
        reason = "Prepating for taking over the world"

        node = self.create_node()

        # Initial setting without the reason
        node = self.conn.baremetal.update_node(node, is_maintenance=True)
        self.assertTrue(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Make sure the change has effect on the remote side.
        node = self.conn.baremetal.get_node(node.id)
        self.assertTrue(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Updating the reason later
        node = self.conn.baremetal.update_node(node, maintenance_reason=reason)
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)

        # Make sure the change has effect on the remote side.
        node = self.conn.baremetal.get_node(node.id)
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)

        # Unsetting maintenance
        node = self.conn.baremetal.update_node(node, is_maintenance=False)
        self.assertFalse(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Make sure the change has effect on the remote side.
        node = self.conn.baremetal.get_node(node.id)
        self.assertFalse(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Initial setting with the reason
        node = self.conn.baremetal.update_node(node, is_maintenance=True,
                                               maintenance_reason=reason)
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)

        # Make sure the change has effect on the remote side.
        node = self.conn.baremetal.get_node(node.id)
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)


class TestBareMetalNodeFields(base.BaseBaremetalTest):

    min_microversion = '1.8'

    def test_node_fields(self):
        self.create_node()
        result = self.conn.baremetal.nodes(
            fields=['uuid', 'name', 'instance_id'])
        for item in result:
            self.assertIsNotNone(item.id)
            self.assertIsNone(item.driver)


class TestBareMetalVif(base.BaseBaremetalTest):

    min_microversion = '1.28'

    def setUp(self):
        super(TestBareMetalVif, self).setUp()
        self.node = self.create_node(network_interface='noop')
        self.vif_id = "200712fc-fdfb-47da-89a6-2d19f76c7618"

    def test_node_vif_attach_detach(self):
        self.conn.baremetal.attach_vif_to_node(self.node, self.vif_id)
        # NOTE(dtantsur): The noop networking driver is completely noop - the
        # VIF list does not return anything of value.
        self.conn.baremetal.list_node_vifs(self.node)
        res = self.conn.baremetal.detach_vif_from_node(self.node, self.vif_id,
                                                       ignore_missing=False)
        self.assertTrue(res)

    def test_node_vif_negative(self):
        uuid = "5c9dcd04-2073-49bc-9618-99ae634d8971"
        self.assertRaises(exceptions.ResourceNotFound,
                          self.conn.baremetal.attach_vif_to_node,
                          uuid, self.vif_id)
        self.assertRaises(exceptions.ResourceNotFound,
                          self.conn.baremetal.list_node_vifs,
                          uuid)
        self.assertRaises(exceptions.ResourceNotFound,
                          self.conn.baremetal.detach_vif_from_node,
                          uuid, self.vif_id, ignore_missing=False)


class TestTraits(base.BaseBaremetalTest):

    min_microversion = '1.37'

    def setUp(self):
        super(TestTraits, self).setUp()
        self.node = self.create_node()

    def test_add_remove_node_trait(self):
        node = self.conn.baremetal.get_node(self.node)
        self.assertEqual([], node.traits)

        self.conn.baremetal.add_node_trait(self.node, 'CUSTOM_FAKE')
        self.assertEqual(['CUSTOM_FAKE'], self.node.traits)
        node = self.conn.baremetal.get_node(self.node)
        self.assertEqual(['CUSTOM_FAKE'], node.traits)

        self.conn.baremetal.add_node_trait(self.node, 'CUSTOM_REAL')
        self.assertEqual(sorted(['CUSTOM_FAKE', 'CUSTOM_REAL']),
                         sorted(self.node.traits))
        node = self.conn.baremetal.get_node(self.node)
        self.assertEqual(sorted(['CUSTOM_FAKE', 'CUSTOM_REAL']),
                         sorted(node.traits))

        self.conn.baremetal.remove_node_trait(node, 'CUSTOM_FAKE',
                                              ignore_missing=False)
        self.assertEqual(['CUSTOM_REAL'], self.node.traits)
        node = self.conn.baremetal.get_node(self.node)
        self.assertEqual(['CUSTOM_REAL'], node.traits)

    def test_set_node_traits(self):
        node = self.conn.baremetal.get_node(self.node)
        self.assertEqual([], node.traits)

        traits1 = ['CUSTOM_FAKE', 'CUSTOM_REAL']
        traits2 = ['CUSTOM_FOOBAR']

        self.conn.baremetal.set_node_traits(self.node, traits1)
        self.assertEqual(sorted(traits1), sorted(self.node.traits))
        node = self.conn.baremetal.get_node(self.node)
        self.assertEqual(sorted(traits1), sorted(node.traits))

        self.conn.baremetal.set_node_traits(self.node, traits2)
        self.assertEqual(['CUSTOM_FOOBAR'], self.node.traits)
        node = self.conn.baremetal.get_node(self.node)
        self.assertEqual(['CUSTOM_FOOBAR'], node.traits)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import _log
from openstack.baremetal.v1 import node as _node
from openstack.baremetal_introspection.v1 import introspection as _introspect
from openstack import exceptions
from openstack import proxy


_logger = _log.setup_logging('openstack')


class Proxy(proxy.Proxy):

    def introspections(self, **query):
        """Retrieve a generator of introspection records.

        :param dict query: Optional query parameters to be sent to restrict
            the records to be returned. Available parameters include:

            * ``fields``: A list containing one or more fields to be returned
              in the response. This may lead to some performance gain
              because other fields of the resource are not refreshed.
            * ``limit``: Requests at most the specified number of items be
              returned from the query.
            * ``marker``: Specifies the ID of the last-seen introspection. Use
              the ``limit`` parameter to make an initial limited request and
              use the ID of the last-seen introspection from the response as
              the ``marker`` value in a subsequent limited request.
            * ``sort_dir``: Sorts the response by the requested sort direction.
              A valid value is ``asc`` (ascending) or ``desc``
              (descending). Default is ``asc``. You can specify multiple
              pairs of sort key and sort direction query parameters. If
              you omit the sort direction in a pair, the API uses the
              natural sorting direction of the server attribute that is
              provided as the ``sort_key``.
            * ``sort_key``: Sorts the response by the this attribute value.
              Default is ``id``. You can specify multiple pairs of sort
              key and sort direction query parameters. If you omit the
              sort direction in a pair, the API uses the natural sorting
              direction of the server attribute that is provided as the
              ``sort_key``.

        :returns: A generator of :class:`~.introspection.Introspection`
            objects
        """
        return _introspect.Introspection.list(self, **query)

    def start_introspection(self, node):
        """Create a new introspection from attributes.

        :param node: The value can be either the name or ID of a node or
            a :class:`~openstack.baremetal.v1.node.Node` instance.

        :returns: :class:`~.introspection.Introspection` instance.
        """
        node = self._get_resource(_node.Node, node)
        return self._create(_introspect.Introspection, id=node.id)

    def get_introspection(self, introspection):
        """Get a specific introspection.

        :param introspection: The value can be the name or ID of an
            introspection (matching bare metal node name or ID) or
            an :class:`~.introspection.Introspection` instance.
        :returns: :class:`~.introspection.Introspection` instance.
        :raises: :class:`~openstack.exceptions.ResourceNotFound` when no
            introspection matching the name or ID could be found.
        """
        return self._get(_introspect.Introspection, introspection)

    def get_introspection_data(self, introspection):
        """Get introspection data.

        :param introspection: The value can be the name or ID of an
            introspection (matching bare metal node name or ID) or
            an :class:`~.introspection.Introspection` instance.
        :returns: introspection data from the most recent successful run.
        :rtype: dict
        """
        res = self._get_resource(_introspect.Introspection, introspection)
        return res.get_data(self)

    def abort_introspection(self, introspection, ignore_missing=True):
        """Abort an introspection.

        Note that the introspection is not aborted immediately, you may use
        `wait_for_introspection` with `ignore_error=True`.

        :param introspection: The value can be the name or ID of an
            introspection (matching bare metal node name or ID) or
            an :class:`~.introspection.Introspection` instance.
        :param bool ignore_missing: When set to ``False``, an exception
            :class:`~openstack.exceptions.ResourceNotFound` will be raised
            when the introspection could not be found. When set to ``True``, no
            exception will be raised when attempting to abort a non-existent
            introspection.
        :returns: nothing
        """
        res = self._get_resource(_introspect.Introspection, introspection)
        try:
            res.abort(self)
        except exceptions.ResourceNotFound:
            if not ignore_missing:
                raise

    def wait_for_introspection(self, introspection, timeout=None,
                               ignore_error=False):
        """Wait for the introspection to finish.

        :param introspection: The value can be the name or ID of an
            introspection (matching bare metal node name or ID) or
            an :class:`~.introspection.Introspection` instance.
        :param timeout: How much (in seconds) to wait for the introspection.
            The value of ``None`` (the default) means no client-side timeout.
        :param ignore_error: If ``True``, this call will raise an exception
            if the introspection reaches the ``error`` state. Otherwise the
            error state is considered successful and the call returns.
        :returns: :class:`~.introspection.Introspection` instance.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if
            introspection fails and ``ignore_error`` is ``False``.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` on timeout.
        """
        res = self._get_resource(_introspect.Introspection, introspection)
        return res.wait(self, timeout=timeout, ignore_error=ignore_error)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.unit import base

from openstack.baremetal.v1 import driver

FAKE = {
    "hosts": [
        "897ab1dad809"
    ],
    "links": [
        {
            "href": "http://127.0.0.1:6385/v1/drivers/agent_ipmitool",
            "rel": "self"
        },
        {
            "href": "http://127.0.0.1:6385/drivers/agent_ipmitool",
            "rel": "bookmark"
        }
    ],
    "name": "agent_ipmitool",
    "properties": [
        {
            "href":
                "http://127.0.0.1:6385/v1/drivers/agent_ipmitool/properties",
            "rel": "self"
        },
        {
            "href": "http://127.0.0.1:6385/drivers/agent_ipmitool/properties",
            "rel": "bookmark"
        }
    ]
}


class TestDriver(base.TestCase):

    def test_basic(self):
        sot = driver.Driver()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('drivers', sot.resources_key)
        self.assertEqual('/drivers', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_instantiate(self):
        sot = driver.Driver(**FAKE)
        self.assertEqual(FAKE['name'], sot.id)
        self.assertEqual(FAKE['name'], sot.name)
        self.assertEqual(FAKE['hosts'], sot.hosts)
        self.assertEqual(FAKE['links'], sot.links)
        self.assertEqual(FAKE['properties'], sot.properties)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import mock

from openstack import exceptions

from openstack.block_storage.v2 import _proxy
from openstack.block_storage.v2 import backup
from openstack.block_storage.v2 import snapshot
from openstack.block_storage.v2 import stats
from openstack.block_storage.v2 import type
from openstack.block_storage.v2 import volume
from openstack.tests.unit import test_proxy_base


class TestVolumeProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super(TestVolumeProxy, self).setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_snapshot_get(self):
        self.verify_get(self.proxy.get_snapshot, snapshot.Snapshot)

    def test_snapshots_detailed(self):
        self.verify_list(self.proxy.snapshots, snapshot.SnapshotDetail,
                         method_kwargs={"details": True, "query": 1},
                         expected_kwargs={"query": 1})

    def test_snapshots_not_detailed(self):
        self.verify_list(self.proxy.snapshots, snapshot.Snapshot,
                         method_kwargs={"details": False, "query": 1},
                         expected_kwargs={"query": 1})

    def test_snapshot_create_attrs(self):
        self.verify_create(self.proxy.create_snapshot, snapshot.Snapshot)

    def test_snapshot_delete(self):
        self.verify_delete(self.proxy.delete_snapshot,
                           snapshot.Snapshot, False)

    def test_snapshot_delete_ignore(self):
        self.verify_delete(self.proxy.delete_snapshot,
                           snapshot.Snapshot, True)

    def test_type_get(self):
        self.verify_get(self.proxy.get_type, type.Type)

    def test_types(self):
        self.verify_list(self.proxy.types, type.Type)

    def test_type_create_attrs(self):
        self.verify_create(self.proxy.create_type, type.Type)

    def test_type_delete(self):
        self.verify_delete(self.proxy.delete_type, type.Type, False)

    def test_type_delete_ignore(self):
        self.verify_delete(self.proxy.delete_type, type.Type, True)

    def test_volume_get(self):
        self.verify_get(self.proxy.get_volume, volume.Volume)

    def test_volumes_detailed(self):
        self.verify_list(self.proxy.volumes, volume.Volume,
                         method_kwargs={"details": True, "query": 1},
                         expected_kwargs={"query": 1,
                                          "base_path": "/volumes/detail"})

    def test_volumes_not_detailed(self):
        self.verify_list(self.proxy.volumes, volume.Volume,
                         method_kwargs={"details": False, "query": 1},
                         expected_kwargs={"query": 1})

    def test_volume_create_attrs(self):
        self.verify_create(self.proxy.create_volume, volume.Volume)

    def test_volume_delete(self):
        self.verify_delete(self.proxy.delete_volume, volume.Volume, False)

    def test_volume_delete_ignore(self):
        self.verify_delete(self.proxy.delete_volume, volume.Volume, True)

    def test_volume_extend(self):
        self._verify("openstack.block_storage.v2.volume.Volume.extend",
                     self.proxy.extend_volume,
                     method_args=["value", "new-size"],
                     expected_args=["new-size"])

    def test_backend_pools(self):
        self.verify_list(self.proxy.backend_pools, stats.Pools)

    def test_backups_detailed(self):
        # NOTE: mock has_service
        self.proxy._connection = mock.Mock()
        self.proxy._connection.has_service = mock.Mock(return_value=True)
        self.verify_list(self.proxy.backups, backup.Backup,
                         method_kwargs={"details": True, "query": 1},
                         expected_kwargs={"query": 1,
                                          "base_path": "/backups/detail"})

    def test_backups_not_detailed(self):
        # NOTE: mock has_service
        self.proxy._connection = mock.Mock()
        self.proxy._connection.has_service = mock.Mock(return_value=True)
        self.verify_list(self.proxy.backups, backup.Backup,
                         method_kwargs={"details": False, "query": 1},
                         expected_kwargs={"query": 1})

    def test_backup_get(self):
        # NOTE: mock has_service
        self.proxy._connection = mock.Mock()
        self.proxy._connection.has_service = mock.Mock(return_value=True)
        self.verify_get(self.proxy.get_backup, backup.Backup)

    def test_backup_delete(self):
        # NOTE: mock has_service
        self.proxy._connection = mock.Mock()
        self.proxy._connection.has_service = mock.Mock(return_value=True)
        self.verify_delete(self.proxy.delete_backup, backup.Backup, False)

    def test_backup_delete_ignore(self):
        # NOTE: mock has_service
        self.proxy._connection = mock.Mock()
        self.proxy._connection.has_service = mock.Mock(return_value=True)
        self.verify_delete(self.proxy.delete_backup, backup.Backup, True)

    def test_backup_create_attrs(self):
        # NOTE: mock has_service
        self.proxy._connection = mock.Mock()
        self.proxy._connection.has_service = mock.Mock(return_value=True)
        self.verify_create(self.proxy.create_backup, backup.Backup)

    def test_backup_restore(self):
        # NOTE: mock has_service
        self.proxy._connection = mock.Mock()
        self.proxy._connection.has_service = mock.Mock(return_value=True)
        self._verify2(
            'openstack.block_storage.v2.backup.Backup.restore',
            self.proxy.restore_backup,
            method_args=['volume_id'],
            method_kwargs={'volume_id': 'vol_id', 'name': 'name'},
            expected_args=[self.proxy],
            expected_kwargs={'volume_id': 'vol_id', 'name': 'name'}
        )

    def test_backup_no_swift(self):
        """Ensure proxy method raises exception if swift is not available
        """
        # NOTE: mock has_service
        self.proxy._connection = mock.Mock()
        self.proxy._connection.has_service = mock.Mock(return_value=False)
        self.assertRaises(
            exceptions.SDKException,
            self.proxy.restore_backup,
            'backup',
            'volume_id',
            'name')
